#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2003  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide soundex calculation
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import unicodedata

# -------------------------------------------------------------------------
#
# constants
#
# -------------------------------------------------------------------------
IGNORE = "HW~!@#$%^&*()_+=-`[]\\|;:'/?.,<>\" \t\f\v"
TABLE = bytes.maketrans(b"ABCDEFGIJKLMNOPQRSTUVXYZ", b"012301202245501262301202")


# -------------------------------------------------------------------------
#
# soundex - returns the soundex value for the specified string
#
# -------------------------------------------------------------------------
def soundex(strval):
    "Return the soundex value to a string argument."

    strval = unicodedata.normalize("NFKD", str(strval.upper().strip())).encode(
        "ASCII", "ignore"
    )
    if not strval:
        return "Z000"
    strval = strval.decode("ASCII", "ignore")
    str2 = strval[0]
    translator = strval.maketrans("", "", IGNORE)
    strval = strval.translate(translator)
    strval = strval.translate(TABLE)
    if not strval:
        return "Z000"
    prev = strval[0]
    for character in strval[1:]:
        if character != prev and character != "0":
            str2 = str2 + character
        prev = character
    # pad with zeros
    str2 = str2 + "0000"
    return str2[:4]


# -------------------------------------------------------------------------
#
# compare - compares the soundex values of two strings
#
# -------------------------------------------------------------------------
def compare(str1, str2):
    "1 if strings are close. 0 otherwise."
    return soundex(str1) == soundex(str2)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gen/mime/__init__.py

try:
    from ._winmime import get_description, get_type, mime_type_is_defined
except:
    from ._pythonmime import get_description, get_type, mime_type_is_defined


def base_type(val):
    return val.split("/")[0]


def is_image_type(val):
    return base_type(val) == "image"


def is_directory(val):
    return base_type(val) == "x-directory"


_invalid_mime_types = ("x-directory", "x-special")


def is_valid_type(val):
    return base_type(val) not in _invalid_mime_types



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007  Donald N. Allingham
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide a simplified database access interface to the Gramps database.
"""
from ..lib import (
    Person,
    Family,
    Event,
    Source,
    Place,
    Citation,
    Media,
    Repository,
    Note,
    Date,
    Tag,
)
from ..errors import HandleError
from ..datehandler import displayer
from ..utils.string import gender as gender_map
from ..utils.db import get_birth_or_fallback, get_death_or_fallback

from ..display.name import displayer as name_displayer
from ..display.place import displayer as place_displayer
from ..lib import EventType
from ..config import config
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Local functions
#
# -------------------------------------------------------------------------
class SimpleAccess:
    """
    Provide a simplified database access system. This system has been designed
    to ease the development of reports.

    The user needs to take care when using this interface. Since it returns real
    objects instead of database references, it can consume a significant amount
    of memory if the user is not careful.

    **Example**

    A typical system of usage would be::

      sa = SimpleAccess(database)

      print "Person        : ", sa.name(person)
      print "Gender        : ", sa.gender(person)
      print "Birth date    : ", sa.birth_date(person)
      print "Birth place   : ", sa.birth_place(person)
      print "Death date    : ", sa.death_date(person)
      print "Death place   : ", sa.death_place(person)
      print "Father        : ", sa.name(sa.father(person))
      print "Mother        : ", sa.name(sa.mother(person))
      print "Spouse        : ", sa.name(sa.spouse(person))
      print "Marriage Type : ", sa.marriage_type(person)
      print "Marriage Date : ", sa.marriage_date(person)
      print "Marriage Place: ", sa.marriage_place(person)
      for child in sa.children(person):
         print "Child         : ", sa.name(child)

      # Print out burial and baptism events
      for event in sa.events( person , [ "Burial", "Baptism" ]):
         print "Event         : ", sa.event_type(event), sa.event_date(event),
         print sa.event_place(event)

    This would produce an output that looks like::

         Person        :  Garner, Lewis Anderson
         Gender        :  male
         Birth date    :  6/21/1855
         Birth place   :  Great Falls, MT
         Death date    :  6/28/1911
         Death place   :  Twin Falls, ID
         Father        :  Garner, Robert W.
         Mother        :  Zielinski, Phoebe Emily
         Spouse        :  Martel, Luella Jacques
         Marriage Type :  Married
         Marriage Date :  4/1/1875
         Marriage Place:  Paragould, AR
         Child         :  Garner, Eugene Stanley
         Child         :  Garner, Jesse V.
         Child         :  Garner, Raymond E.
         Child         :  Garner, Jennie S.
         Child         :  Garner, Walter E.
         Child         :  Garner, Daniel Webster
         Child         :  Garner, Bertha P.
         Child         :  Garner, Elizabeth
         Event         :  Burial 7/1/1911 Twin Falls, ID

    One can also alternatively supply a handle rather than a person.
    """

    def __init__(self, dbase):
        """
        Initialize the SimpleAccess object with the database that will be used.

        :param dbase: Gramps database object
        :type dbase: DbBase
        """
        self.dbase = dbase

    def name(self, person):
        """
        Return the name of the person, or and empty string if the person is None

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns the name of the person based of the program preferences
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)
        if person:
            return name_displayer.display(person)
        else:
            return ""

    def surname(self, person):
        """
        Return the name of the person, or and empty string if the person is None

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns the name of the person based of the program preferences
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)
        if person:
            surname = person.get_primary_name().get_surname()
            return surname or config.get("preferences.no-surname-text")
        else:
            return ""

    def first_name(self, person):
        """
        Return the first name of the person, or and empty string if the person
        is None

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns the first name of the person based of the program
                 preferences
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)
        if person:
            return person.get_primary_name().get_first_name()
        else:
            return ""

    def gid(self, obj):
        """
        Return the Gramps ID of the person or family

        :param obj: Person or Family object
        :type obj: :py:class:`.Person` or :py:class:`.Family`
        :return: Returns the Gramps ID value of the person or family
        :rtype: unicode
        """
        if obj:
            return obj.get_gramps_id()
        else:
            return ""

    def gender(self, person):
        """
        Return a string representing the gender of the person

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns a string indentifying the person's gender
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)
        if person:
            return gender_map[person.get_gender()]
        return ""

    def __parent(self, person, func):
        """
        Return a person associated as a parent of the person

        :param person: Person object
        :type person: :py:class:`.Person`
        :param func: function used to extract the appropriate parent
        :type func: function
        :return: mother or father of the associated person
        :rtype: :py:class:`.Person`
        """
        assert person is None or isinstance(person, Person)

        if person:
            parent_handle_list = person.get_parent_family_handle_list()
            if parent_handle_list:
                family_id = parent_handle_list[0]
                family = self.dbase.get_family_from_handle(family_id)
                return self.__family_parent(family, func)
        return None

    def __family_parent(self, family, func):
        """
        Return a person associated as a parent of the family

        :param family: Family object
        :type family: :py:class:`.Family`
        :param func: function used to extract the appropriate parent
        :type func: function
        :return: mother or father of the associated family
        :rtype: :py:class:`.Family`
        """
        assert family is None or isinstance(family, Family)

        if family:
            handle = func(family)
            if handle:
                return self.dbase.get_person_from_handle(handle)
        return None

    def __event_date(self, person, func):
        """
        Return a string describing the date associated with the person

        :param person: Person object
        :type person: :py:class:`.Person`
        :param func: function used to extract the associated date information
        :type func: function
        :return: Returns a string describing the date
        :rtype: unicode
        """
        assert person is None or isinstance(person, Person)

        if person:
            ref = func(person)
            if ref:
                event_handle = ref.get_reference_handle()
                if event_handle:
                    event = self.dbase.get_event_from_handle(event_handle)
                    date_obj = event.get_date_object()
                    if date_obj:
                        return displayer.display(date_obj)
        return ""

    def __event_date_obj(self, person, func):
        """
        Return the date associated with the person

        :param person: Person object
        :type person: :py:class:`.Person`
        :param func: function used to extract the associated date information
        :type func: function
        :return: Returns the date
        :rtype: l{gen.lib.Date}
        """
        assert person is None or isinstance(person, Person)

        if person:
            ref = func(person)
            if ref:
                event_handle = ref.get_reference_handle()
                if event_handle:
                    event = self.dbase.get_event_from_handle(event_handle)
                    date_obj = event.get_date_object()
                    if date_obj:
                        return date_obj
                    else:
                        return Date()
        return Date()

    def __event_place(self, person, func):
        """
        Return a string describing the place associated with the person

        :param person: Person object
        :type person: :py:class:`.Person`
        :param func: function used to extract the associated place information
        :type func: function
        :return: Returns a string describing the place
        :rtype: unicode
        """
        assert person is None or isinstance(person, Person)

        if person:
            ref = func(person)
            if ref:
                event_handle = ref.get_reference_handle()
                if event_handle:
                    event = self.dbase.get_event_from_handle(event_handle)
                    return place_displayer.display_event(self.dbase, event)
        return ""

    def spouse(self, person):
        """
        Return the primary spouse of the person

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: The spouse identified as the person's primary spouse
        :rtype: :py:class:`.Person`
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)

        if person:
            family_handle_list = person.get_family_handle_list()
            if family_handle_list:
                family_id = family_handle_list[0]
                family = self.dbase.get_family_from_handle(family_id)
                if family.get_father_handle() == person.get_handle():
                    person_handle = family.get_mother_handle()
                else:
                    person_handle = family.get_father_handle()
                if person_handle:
                    return self.dbase.get_person_from_handle(person_handle)
        return None

    def marriage_type(self, person):
        """
        Return a string describing the relationship between the person and
        his/per primary spouse.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns a string describing the relationship between the
                 person and his/per primary spouse.
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)

        if person:
            family_handle_list = person.get_family_handle_list()
            if family_handle_list:
                family_id = family_handle_list[0]
                family = self.dbase.get_family_from_handle(family_id)
                if family:
                    return str(family.get_relationship())
        return ""

    def marriage_place(self, person):
        """
        Return a string describing the place where the person and his/her spouse
        where married.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns a string describing the place where the person and
                 his/her spouse where married.
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)

        if person:
            family_handle_list = person.get_family_handle_list()
            if family_handle_list:
                family_id = family_handle_list[0]
                family = self.dbase.get_family_from_handle(family_id)
                if family:
                    reflist = family.get_event_ref_list()
                    if reflist:
                        # 'evnt' should be renamed to 'event'?
                        elist = [
                            self.dbase.get_event_from_handle(ref.ref) for ref in reflist
                        ]
                        events = [
                            evnt for evnt in elist if evnt.type == EventType.MARRIAGE
                        ]
                        if events:
                            return place_displayer.display_event(self.dbase, events[0])
        return ""

    def marriage_date(self, person):
        """
        Return a string indicating the date when the person and his/her spouse
        where married.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns a string indicicating the date when the person and
                 his/her spouse where married.
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)

        if person:
            family_handle_list = person.get_family_handle_list()
            if family_handle_list:
                family_id = family_handle_list[0]
                family = self.dbase.get_family_from_handle(family_id)
                if family:
                    reflist = family.get_event_ref_list()
                    if reflist:
                        elist = [
                            self.dbase.get_event_from_handle(ref.ref) for ref in reflist
                        ]
                        events = [
                            evnt for evnt in elist if evnt.type == EventType.MARRIAGE
                        ]
                        if events:
                            date_obj = events[0].get_date_object()
                            if date_obj:
                                return displayer.display(date_obj)
        return ""

    def children(self, obj):
        """
        Return a list of the children as the children of the primary spouse.

        :param obj: Person or Family object
        :type obj: :py:class:`.Person` or :py:class:`.Family`
        :return: Returns a list of :py:class:`.Person` objects representing the
                 children
        :rtype: list
        """
        assert obj is None or isinstance(obj, (Person, Family))

        if isinstance(obj, Person):
            family_handle_list = obj.get_family_handle_list()
            if family_handle_list:
                family_id = family_handle_list[0]
                family = self.dbase.get_family_from_handle(family_id)

                return [
                    self.dbase.get_person_from_handle(hndl.ref)
                    for hndl in family.get_child_ref_list()
                ]
        elif isinstance(obj, Family):
            return [
                self.dbase.get_person_from_handle(hndl.ref)
                for hndl in obj.get_child_ref_list()
            ]
        return []

    def father(self, obj):
        """
        Return the primary father of the person or the father of the associated
        family.

        :param obj: Person or Family object
        :type obj: :py:class:`.Person` or :py:class:`.Family`
        :return: The father in the person's primary family or the father of the
                 family
        :rtype: :py:class:`.Person`
        """
        if isinstance(obj, Person):
            return self.__parent(obj, Family.get_father_handle)
        elif isinstance(obj, Family):
            return self.__family_parent(obj, Family.get_father_handle)
        else:
            return None

    def mother(self, obj):
        """
        Returns the primary mother of the person or the mother of the associated
        family.

        :param obj: Person object
        :type obj: :py:class:`.Person` or :py:class:`.Family`
        :return: The mother in the person's primary family or the mother of the
                 family
        :rtype: :py:class:`.Person`
        """
        if isinstance(obj, Person):
            return self.__parent(obj, Family.get_mother_handle)
        elif isinstance(obj, Family):
            return self.__family_parent(obj, Family.get_mother_handle)
        else:
            return None

    def birth_date(self, person):
        """
        Return a string indicating the date when the person's birth.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns a string indicating the date when the person's birth.
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        return self.__event_date(person, Person.get_birth_ref)

    def birth_date_obj(self, person):
        """
        Return the date when the person's birth.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns the date when the person's birth.
        :rtype: :py:class:`.Date`
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        return self.__event_date_obj(person, Person.get_birth_ref)

    def birth_or_fallback(self, person, get_event=False):
        """
        Return the date of the person's birth or fallback event.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns the date when the person's birth or fallback.
        :rtype: :py:class:`.Date`
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        event = get_birth_or_fallback(self.dbase, person, "<i>%s</i>")
        if get_event:
            return event
        elif event:
            return event.date
        else:
            return Date()

    def birth_place(self, person):
        """
        Return a string indicating the place of the person's birth.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns a string indicating the place of the person's birth.
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        return self.__event_place(person, Person.get_birth_ref)

    def death_date(self, person):
        """
        Return a string indicating the date when the person's death.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns a string indicating the date when the person's death.
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        return self.__event_date(person, Person.get_death_ref)

    def death_date_obj(self, person):
        """
        Return the date when the person's death.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns the date when the person's death.
        :rtype: :py:class:`.Date`
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        return self.__event_date_obj(person, Person.get_death_ref)

    def death_or_fallback(self, person, get_event=False):
        """
        Return the date of the person's death or fallback event.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns the date of the person's death or fallback.
        :rtype: :py:class:`.Date`
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        event = get_death_or_fallback(self.dbase, person, "<i>%s</i>")
        if get_event:
            return event
        elif event:
            return event.date
        else:
            return Date()

    def death_place(self, person):
        """
        Return a string indicating the place of the person's death.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: Returns a string indicating the place of the person's death.
        :rtype: unicode
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        return self.__event_place(person, Person.get_death_ref)

    def event_place(self, event):
        """
        Return a string indicating the place of the event

        :param event: Event object
        :type event: :py:class:`.Event`
        :return: Returns a string indicating the place of the event
        :rtype: unicode
        """
        assert event is None or isinstance(event, Event)

        if event:
            return place_displayer.display_event(self.dbase, event)
        else:
            return ""

    def date_string(self, date_obj):
        """
        Return a string representation a date_obj

        :param date_obj: Date object
        :type date_obj: :py:class:`.Date`
        :return: Returns a string representation a date_obj
        :rtype: unicode
        """
        if date_obj:
            return displayer.display(date_obj)
        else:
            return ""

    def event_date(self, event):
        """
        Return a string indicating the date of the event

        :param event: Event object
        :type event: :py:class:`.Event`
        :return: Returns a string indicating the date of the event
        :rtype: unicode
        """
        assert event is None or isinstance(event, Event)
        date_obj = event.get_date_object()
        if date_obj:
            return displayer.display(date_obj)
        else:
            return ""

    def event_date_obj(self, event):
        """
        Return a string indicating the date of the event

        :param event: Event object
        :type event: :py:class:`.Event`
        :return: Returns a string indicating the date of the event
        :rtype: unicode
        """
        assert event is None or isinstance(event, Event)
        if event:
            return event.get_date_object()

    def event_type(self, event):
        """
        Return a string indicating the type of the event

        :param event: Event object
        :type event: :py:class:`.Event`
        :return: Returns a string indicating the type of the event
        :rtype: unicode
        """
        assert event is None or isinstance(event, Event)
        if event:
            return str(event.get_type())
        else:
            return ""

    def events(self, obj, restrict=None):
        """
        Return a list of events associated with the object. This object
        can be either a :py:class:`.Person` or :py:class:`.Family`.

        :param obj: Person or Family
        :type obj: :py:class:`.Person` or :py:class:`.Family`
        :param restrict: Optional list of strings that will limit the types
                         of events to those of the specified types.
        :type restrict: list
        :return: list of events associated with the object
        :rtype: list
        """
        assert obj is None or isinstance(obj, (Person, Family))
        assert isinstance(restrict, list) or restrict is None

        if obj:
            event_handles = [ref.ref for ref in obj.get_event_ref_list()]
            events = [self.dbase.get_event_from_handle(h) for h in event_handles]
            if restrict:
                restrict = [r.lower() for r in restrict]
                events = [
                    event
                    for event in events
                    if str(event.get_type()).lower() in restrict
                ]
            return events
        else:
            return []

    def sources(self, obj):
        """
        Return a list of events associated with the object. This object
        can be either a :py:class:`.Person` or :py:class:`.Family`.

        :param obj: Person or Family
        :type obj: :py:class:`.Person` or :py:class:`.Family`
        :return: list of events associated with the object
        :rtype: list
        """
        assert obj is None or isinstance(obj, (Person, Family, Event))

        if obj:
            handles = [ref.ref for ref in obj.get_source_references()]
            return list(map(self.dbase.get_source_from_handle, handles))
        else:
            return []

    def parent_in(self, person):
        """
        Return a list of families in which the person is listed as a parent.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: list of :py:class:`.Family` objects in which the person is
                 listed as a parent.
        :rtype: list
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)

        if person:
            return [
                self.dbase.get_family_from_handle(handle)
                for handle in person.get_family_handle_list()
            ]
        return []

    def child_in(self, person):
        """
        Return a list of families in which the person is listed as a child.

        :param person: Person object
        :type person: :py:class:`.Person`
        :return: list of :py:class:`.Family` objects in which the person is
                 listed as a child.
        :rtype: list
        """
        if isinstance(person, str):
            person = self.dbase.get_person_from_handle(person)
        assert person is None or isinstance(person, Person)

        if person:
            return [
                self.dbase.get_family_from_handle(handle)
                for handle in person.get_parent_family_handle_list()
            ]
        return []

    def __all_objects(self, gen_cursor, get_object):
        """
        Return a all the objects of a particular type in the database, one
        at a time as an iterator. The user can treat this just like a list.

        :return: list of objects of a particular type in the database
        :rtype: list
        """

        with gen_cursor() as cursor:
            for key, data in cursor:
                yield get_object(key)

    def all_people(self):
        """
        Return a all the people in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

           for person in sa.all_people():
              sa.print(person)

        :return: list of people in the database
        :rtype: list
        """

        with self.dbase.get_person_cursor() as cursor:
            slist = sorted(
                (
                    (
                        data.primary_name.surname_list[0].surname
                        if data.primary_name.surname_list
                        else ""
                    ),
                    key,
                )
                for key, data in cursor
            )

        for info in slist:
            obj = self.dbase.get_person_from_handle(info[1])
            yield obj

    def all_families(self):
        """
        Return all the families in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

           for person in sa.all_families():
              sa.print(sa.father(person))

        :return: list of families in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_family_cursor, self.dbase.get_family_from_handle
        )

    def all_events(self):
        """
        Return all the events in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

           for person in sa.all_events():
              sa.print(sa.event_place(event))

        :return: list of events in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_event_cursor, self.dbase.get_event_from_handle
        )

    def all_sources(self):
        """
        Return all the sources in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

        :return: list of sources in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_source_cursor, self.dbase.get_source_from_handle
        )

    def all_citations(self):
        """
        Return all the citations in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

        :return: list of citations in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_citation_cursor, self.dbase.get_citation_from_handle
        )

    def all_repositories(self):
        """
        Return all the repositories in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

        :return: list of repositories in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_repository_cursor, self.dbase.get_repository_from_handle
        )

    def all_media(self):
        """
        Return all the media in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

        :return: list of media in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_media_cursor, self.dbase.get_media_from_handle
        )

    def all_places(self):
        """
        Return all the places in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

        :return: list of places in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_place_cursor, self.dbase.get_place_from_handle
        )

    def all_notes(self):
        """
        Return all the notes in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

        :return: list of notes in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_note_cursor, self.dbase.get_note_from_handle
        )

    def all_tags(self):
        """
        Return all the tags in the database, one at a time as an iterator.
        The user can treat this just like a list. For example::

        :return: list of tags in the database
        :rtype: list
        """
        return self.__all_objects(
            self.dbase.get_tag_cursor, self.dbase.get_tag_from_handle
        )

    def title(self, source):
        """
        Return the title of the source.

        :param source: Source object
        :type source: :py:class:`.Source`
        :return: title of the source
        :rtype: unicode
        """
        assert source is None or isinstance(source, Source)
        if source:
            return source.get_title()
        return ""

    def page(self, citation):
        """
        Return the page of the citation.

        :param citation: Source object
        :type citation: :py:class:`Citation <.lib.citation.Citation>`
        :return: title of the citation
        :rtype: unicode
        """
        assert citation is None or isinstance(citation, Citation)
        if citation:
            return citation.get_page()
        return ""

    def author(self, source):
        """
        Return the author of the source.

        :param source: Source object
        :type source: :py:class:`.Source`
        :return: author of the source
        :rtype: unicode
        """
        assert source is None or isinstance(source, Source)
        if source:
            return source.get_author()
        return ""

    def person(self, handle):
        assert isinstance(handle, str)
        return self.dbase.get_person_from_handle(handle)

    def event(self, handle):
        assert isinstance(handle, str)
        return self.dbase.get_event_from_handle(handle)

    def family(self, handle):
        assert isinstance(handle, str)
        return self.dbase.get_family_from_handle(handle)

    def display(self, object_class, prop, value):
        """
        Given a object_class, prop, and value return a display string
        describing object.

        :param object_class: "Person", "Source", etc.
        :param prop: "gramps_id", or "handle"
        :param value: gramps_id or handle.
        """
        func = self.dbase.method("get_%s_from_%s", object_class, prop)
        if func:
            try:
                obj = func(value)
            except HandleError:
                # Deals with deleted objects referenced in Note Links
                obj = None
            if obj:
                if isinstance(obj, Person):
                    return "%s: %s [%s]" % (
                        _(object_class),
                        self.name(obj),
                        self.gid(obj),
                    )
                elif isinstance(obj, Event):
                    return "%s: %s [%s]" % (
                        _(object_class),
                        self.event_type(obj),
                        self.gid(obj),
                    )
                elif isinstance(obj, Family):
                    return "%s: %s/%s [%s]" % (
                        _(object_class),
                        self.name(self.mother(obj)),
                        self.name(self.father(obj)),
                        self.gid(obj),
                    )
                elif isinstance(obj, Media):
                    return "%s: %s [%s]" % (_(object_class), obj.desc, self.gid(obj))
                elif isinstance(obj, Source):
                    return "%s: %s [%s]" % (
                        _(object_class),
                        self.title(obj),
                        self.gid(obj),
                    )
                elif isinstance(obj, Citation):
                    return "%s: [%s]" % (_(object_class), self.gid(obj))
                elif isinstance(obj, Place):
                    place_title = place_displayer.display(self.dbase, obj)
                    return "%s: %s [%s]" % (_(object_class), place_title, self.gid(obj))
                elif isinstance(obj, Repository):
                    return "%s: %s [%s]" % (_(object_class), obj.type, self.gid(obj))
                elif isinstance(obj, Note):
                    return "%s: %s [%s]" % (_(object_class), obj.type, self.gid(obj))
                elif isinstance(obj, Tag):
                    return "%s: [%s]" % (_(object_class), obj.name)
                else:
                    return "Error: incorrect object class in display: '%s'" % type(obj)
            else:
                return "Error: missing object"
        else:
            return "Error: invalid object class in display: '%s'" % object_class

    def describe(self, obj, prop=None, value=None):
        """
        Given a object, return a string describing the object.
        """
        if prop and value:
            func = self.dbase.method("get_%s_from_%s", object_class, prop)
            if func:
                obj = func(value)
        if isinstance(obj, Person):
            return "%s [%s]" % (self.name(obj), self.gid(obj))
        elif isinstance(obj, Event):
            return "%s [%s]" % (self.event_type(obj), self.gid(obj))
        elif isinstance(obj, Family):
            return "%s/%s [%s]" % (
                self.name(self.mother(obj)),
                self.name(self.father(obj)),
                self.gid(obj),
            )
        elif isinstance(obj, Media):
            return "%s [%s]" % (obj.desc, self.gid(obj))
        elif isinstance(obj, Source):
            return "%s [%s]" % (self.title(obj), self.gid(obj))
        elif isinstance(obj, Citation):
            return "[%s]" % (self.gid(obj))
        elif isinstance(obj, Place):
            place_title = place_displayer.display(self.dbase, obj)
            return "%s [%s]" % (place_title, self.gid(obj))
        elif isinstance(obj, Repository):
            return "%s [%s]" % (obj.type, self.gid(obj))
        elif isinstance(obj, Note):
            return "%s [%s]" % (obj.type, self.gid(obj))
        elif isinstance(obj, Tag):
            return "[%s]" % (obj.name)
        else:
            return "Error: incorrect object class in describe: '%s'" % type(obj)

    def get_link(self, object_class, prop, value):
        """
        Given a object_class, prop, and value return the object.

        :param object_class: "Person", "Source", etc.
        :param prop: "gramps_id", or "handle"
        :param value: gramps_id or handle.
        """
        func = self.dbase.method("get_%s_from_%s", object_class, prop)
        if func:
            return func(value)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Serbian version by Vlada Perić <vlada.peric@gmail.com>, 2009.
# Based on the Croatian DateHandler by Josip

"""
Serbian-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Serbian parser
#
# -------------------------------------------------------------------------
class DateParserSR(DateParser):
    """
    Converts a text string into a Date object
    """

    month_to_int = DateParser.month_to_int

    month_to_int["januar"] = 1
    month_to_int["januara"] = 1
    month_to_int["jan"] = 1
    month_to_int["јан"] = 1
    month_to_int["јануара"] = 1
    month_to_int["јануар"] = 1
    month_to_int["i"] = 1

    month_to_int["februar"] = 2
    month_to_int["februara"] = 2
    month_to_int["feb"] = 2
    month_to_int["феб"] = 2
    month_to_int["фебруар"] = 2
    month_to_int["фебруара"] = 2
    month_to_int["ii"] = 2

    month_to_int["mart"] = 3
    month_to_int["marta"] = 3
    month_to_int["mar"] = 3
    month_to_int["мар"] = 3
    month_to_int["март"] = 3
    month_to_int["марта"] = 3
    month_to_int["iii"] = 3

    month_to_int["april"] = 4
    month_to_int["aprila"] = 4
    month_to_int["apr"] = 4
    month_to_int["апр"] = 4
    month_to_int["април"] = 4
    month_to_int["априла"] = 4
    month_to_int["iv"] = 4

    month_to_int["maj"] = 5
    month_to_int["maja"] = 5
    month_to_int["мај"] = 5
    month_to_int["маја"] = 5
    month_to_int["v"] = 5

    month_to_int["jun"] = 6
    month_to_int["juna"] = 6
    month_to_int["јун"] = 6
    month_to_int["јуна"] = 6
    month_to_int["vi"] = 6

    month_to_int["jul"] = 7
    month_to_int["jula"] = 7
    month_to_int["јул"] = 7
    month_to_int["јула"] = 7
    month_to_int["vii"] = 7

    month_to_int["avgust"] = 8
    month_to_int["avgusta"] = 8
    month_to_int["avg"] = 8
    month_to_int["авг"] = 8
    month_to_int["август"] = 8
    month_to_int["августа"] = 8
    month_to_int["viii"] = 8

    month_to_int["septembar"] = 9
    month_to_int["septembra"] = 9
    month_to_int["sep"] = 9
    month_to_int["сеп"] = 9
    month_to_int["септембар"] = 9
    month_to_int["септембра"] = 9
    month_to_int["ix"] = 9

    month_to_int["oktobar"] = 10
    month_to_int["oktobra"] = 10
    month_to_int["okt"] = 10
    month_to_int["окт"] = 10
    month_to_int["октобар"] = 10
    month_to_int["октобра"] = 10
    month_to_int["x"] = 10

    month_to_int["novembar"] = 11
    month_to_int["novembra"] = 11
    month_to_int["nov"] = 11
    month_to_int["нов"] = 11
    month_to_int["новембар"] = 11
    month_to_int["новембра"] = 11
    month_to_int["xi"] = 11

    month_to_int["decembar"] = 12
    month_to_int["decembra"] = 12
    month_to_int["dec"] = 12
    month_to_int["дец"] = 12
    month_to_int["децембар"] = 12
    month_to_int["децембра"] = 12
    month_to_int["xii"] = 12

    modifier_to_int = {
        "pre": Date.MOD_BEFORE,
        "posle": Date.MOD_AFTER,
        "oko": Date.MOD_ABOUT,
        "cca": Date.MOD_ABOUT,
        "пре": Date.MOD_BEFORE,
        "после": Date.MOD_AFTER,
        "око": Date.MOD_ABOUT,
        "from": Date.MOD_FROM,
        "to": Date.MOD_TO,
    }

    calendar_to_int = {
        "gregorijanski": Date.CAL_GREGORIAN,
        "greg.": Date.CAL_GREGORIAN,
        "julijanski": Date.CAL_JULIAN,
        "jul.": Date.CAL_JULIAN,
        "hebrejski": Date.CAL_HEBREW,
        "hebr.": Date.CAL_HEBREW,
        "islamski": Date.CAL_ISLAMIC,
        "isl.": Date.CAL_ISLAMIC,
        "francuski republikanski": Date.CAL_FRENCH,
        "franc.": Date.CAL_FRENCH,
        "persijski": Date.CAL_PERSIAN,
        "pers. ": Date.CAL_PERSIAN,
        "švedski": Date.CAL_SWEDISH,
        "šv.": Date.CAL_SWEDISH,
        "грегоријански": Date.CAL_GREGORIAN,
        "грег.": Date.CAL_GREGORIAN,
        "јулијански": Date.CAL_JULIAN,
        "јул.": Date.CAL_JULIAN,
        "хебрејски": Date.CAL_HEBREW,
        "хебр.": Date.CAL_HEBREW,
        "исламски": Date.CAL_ISLAMIC,
        "исл.": Date.CAL_ISLAMIC,
        "француски републикански": Date.CAL_FRENCH,
        "франц.": Date.CAL_FRENCH,
        "персијски": Date.CAL_PERSIAN,
        "перс. ": Date.CAL_PERSIAN,
        "шведски": Date.CAL_SWEDISH,
        "шв": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "procenjeno": Date.QUAL_ESTIMATED,
        "pro.": Date.QUAL_ESTIMATED,
        "izračunato": Date.QUAL_CALCULATED,
        "izr.": Date.QUAL_CALCULATED,
        "процењено": Date.QUAL_ESTIMATED,
        "про.": Date.QUAL_ESTIMATED,
        "приближно": Date.QUAL_ESTIMATED,
        "израчунато": Date.QUAL_CALCULATED,
        "изр.": Date.QUAL_CALCULATED,
        "прорачунато": Date.QUAL_CALCULATED,
    }

    bce = [
        "пре нове ере",
        "пре Христа",
        "п.н.е." "pre nove ere",
        "pre Hrista",
        "p.n.e.",
    ] + DateParser.bce

    def init_strings(self):
        """
        compiles regular expression strings for matching dates
        """
        DateParser.init_strings(self)
        # match 'Day. MONTH year.' format with or without dots
        self._text2 = re.compile(
            r"(\d+)?\.?\s*?%s\s*((\d+)(/\d+)?)?\.?\s*$" % self._mon_str, re.IGNORECASE
        )

        # match Day.Month.Year.
        self._numeric = re.compile(r"((\d+)[/\. ])?\s*((\d+)[/\.])?\s*(\d+)\.?$")

        _span_1 = ["od", "од", "из"]
        _span_2 = ["do", "до"]
        _range_1 = ["između", "између"]
        _range_2 = ["i", "и"]
        self._span = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Serbian display
#
# -------------------------------------------------------------------------
class DateDisplaySR_Base(DateDisplay):
    """
    Serbian (base) date display class
    """

    roman_months = (
        "",
        "I",
        "II",
        "III",
        "IV",
        "V",
        "VI",
        "VII",
        "VIII",
        "IX",
        "X",
        "XI",
        "XII",
    )

    def _display_gregorian(self, date_val, **kwargs):
        """
        display gregorian calendar date in different format
        """
        # this must agree with its locale-specific "formats" definition
        year = self._slash_year(date_val[2], date_val[3])
        if self.format == 0:
            return self.display_iso(date_val)
        elif self.format == 1:
            # day.month_number.year.
            if date_val[3]:
                return self.display_iso(date_val)
            else:
                if date_val[0] == 0 and date_val[1] == 0:
                    value = str(date_val[2])
                else:
                    value = self.dhformat.replace("%m", str(date_val[1]))
                    value = value.replace("%d", str(date_val[0]))
                    value = value.replace("%Y", str(abs(date_val[2])))
                    # some locale magic already provides the right separator
                    # value = value.replace('/', '.')
        elif self.format == 2:
            # day. month_abbreviation year.
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = "%s." % year
                else:
                    value = "%s %s." % (self.short_months[date_val[1]], year)
            else:
                value = "%d. %s %s." % (
                    date_val[0],
                    self.short_months[date_val[1]],
                    year,
                )
        elif self.format == 3:
            # day. month_name year.
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = "%s." % year
                else:
                    value = "%s %s." % (self.long_months[date_val[1]], year)
            else:
                value = "%d. %s %s." % (
                    date_val[0],
                    self.long_months[date_val[1]],
                    year,
                )
        else:
            # day. Roman_number_month year.
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = "%s." % year
                else:
                    value = "%s %s." % (self.roman_months[date_val[1]], year)
            else:
                value = "%d. %s %s." % (
                    date_val[0],
                    self.roman_months[date_val[1]],
                    year,
                )
        if date_val[2] < 0:
            return self._bce_str % value
        else:
            return value

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]
        span1 = self._span1
        span2 = self._span2
        range1 = self._range1
        range2 = self._range2

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d_1 = self.display_cal[cal](start)
            d_2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, span1, d_1, span2, d_2, scal)
        elif mod == Date.MOD_RANGE:
            d_1 = self.display_cal[cal](start)
            d_2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, range1, d_1, range2, d_2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)


class DateDisplaySR_Latin(DateDisplaySR_Base):
    """
    Serbian (Latin) date display class
    """

    long_months = (
        "",
        "januara",
        "februara",
        "marta",
        "aprila",
        "maja",
        "juna",
        "jula",
        "avgusta",
        "septembra",
        "oktobra",
        "novembra",
        "decembra",
    )

    short_months = (
        "",
        "jan",
        "feb",
        "mar",
        "apr",
        "maj",
        "jun",
        "jul",
        "avg",
        "sep",
        "okt",
        "nov",
        "dec",
    )

    calendar = (
        "",
        "julijanski",
        "hebrejski",
        "francuski republikanski",
        "persijski",
        "islamski",
        "švedski",
    )

    _mod_str = ("", "pre ", "posle ", "oko ", "", "", "")

    _qual_str = ("", "procenjeno ", "izračunato ")

    _bce_str = "%s p.n.e."

    formats = (
        "GGGG-MM-DD (ISO-8601)",
        "Numerički (DD.MM.GGGG.)",
        "D. MMM GGGG.",
        "D. Mesec GGGG.",
        "D. Rb GGGG.",
    )
    # this definition must agree with its "_display_gregorian" method

    _span1 = "od"
    _span2 = "do"
    _range1 = "između"
    _range2 = "i"


class DateDisplaySR_Cyrillic(DateDisplaySR_Base):
    """
    Serbian (Cyrillic) date display class
    """

    long_months = (
        "",
        "Јануар",
        "Фебруар",
        "Март",
        "Април",
        "Мај",
        "Јуне",
        "Јули",
        "Аугуст",
        "Септембар",
        "Оцтобер",
        "Новембер",
        "Децембар",
    )

    short_months = (
        "",
        "Јан",
        "Феб",
        "Мар",
        "Апр",
        "Мај",
        "Јун",
        "Јул",
        "Авг",
        "Сеп",
        "Окт",
        "Нов",
        "Дец",
    )

    calendar = (
        "",
        "Јулиан",
        "хебрејски",
        "француски републиканац",
        "Персиан",
        "исламски",
        "шведски",
    )

    _mod_str = ("", "пре ", "после ", "око ", "", "", "")

    _qual_str = ("", "процењено ", "израчунато ")

    _bce_str = "%s п.н.е."

    formats = (
        "ГГГГ-ММ-ДД (ISO-8601)",
        "Нумеричка (ДД.ММ.ГГГГ.)",
        "Д. МММ ГГГГ.",
        "Д. Месец ГГГГ.",
        "Д. Rb ГГГГ.",
    )
    # this definition must agree with its "_display_gregorian" method

    _span1 = "из"
    _span2 = "до"
    _range1 = "између"
    _range2 = "и"


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    (
        "sr_RS.utf8@latin",
        "srpski",
        "Srpski",
        "sr_Latn",
        "sr_Latn_RS",
        "sr_RS@latin",
        ("%d.%m.%Y.",),
    ),
    DateParserSR,
    DateDisplaySR_Latin,
)
register_datehandler(
    (
        "sr_RS",
        "sr",
        "sr_Cyrl",
        "sr_Cyrl_RS",
        "српски",
        "Српски",
        "serbian",
        ("%d.%m.%Y.",),
    ),
    DateParserSR,
    DateDisplaySR_Cyrillic,
)



# encoding:utf-8
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
# Copyright (C) 2011 Nick Hall
# Copyright (C) 2011 Tim G L Lyons
# Copyright (C) 2020 Matthias Kemmer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import register, STABLE, UNSTABLE, GRAMPLET
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"
GRAMPLET_HELP = _("Gramps_6.0_Wiki_Manual_-_Gramplets#Gramplet_List")
DEBUG_HELP = _("Gramps_6.0_Wiki_Manual_-_Tools#Debug")

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------

from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# ------------------------------------------------------------------------
#
# Register Gramplet
#
# ------------------------------------------------------------------------
register(
    GRAMPLET,
    id="Age on Date",
    name=_("Age on Date"),
    description=_("Gramplet showing ages of living people on a specific date"),
    version="2.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="ageondategramplet.py",
    height=200,
    gramplet="AgeOnDateGramplet",
    gramplet_title=_("Age on Date"),
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Age Stats",
    name=_("Age Stats"),
    description=_("Gramplet showing graphs of various ages"),
    status=STABLE,
    fname="agestats.py",
    height=100,
    expand=True,
    gramplet="AgeStatsGramplet",
    gramplet_title=_("Age Stats"),
    detached_width=600,
    detached_height=450,
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Calendar",
    name=_("Calendar"),
    description=_("Gramplet showing calendar and events on specific dates in history"),
    status=STABLE,
    fname="calendargramplet.py",
    height=200,
    gramplet="CalendarGramplet",
    gramplet_title=_("Calendar"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Descendant",
    name=_("Descendants"),
    description=_("Gramplet showing active person's descendants"),
    status=STABLE,
    fname="descendant.py",
    height=100,
    expand=True,
    gramplet="Descendant",
    gramplet_title=_("Descendants"),
    detached_width=500,
    detached_height=500,
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Ancestor",
    name=_("Ancestors"),
    description=_("Gramplet showing active person's ancestors"),
    status=STABLE,
    fname="ancestor.py",
    height=100,
    expand=True,
    gramplet="Ancestor",
    gramplet_title=_("Ancestors"),
    detached_width=500,
    detached_height=500,
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Fan Chart",
    name=_("Fan Chart"),
    description=_("Gramplet showing active person's direct ancestors as a fanchart"),
    status=STABLE,
    fname="fanchartgramplet.py",
    height=430,
    expand=True,
    gramplet="FanChartGramplet",
    detached_height=550,
    detached_width=475,
    gramplet_title=_("Fan Chart"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Descendant Fan Chart",
    name=_("Descendant Fan Chart"),
    description=_("Gramplet showing active person's direct descendants as a fanchart"),
    status=STABLE,
    fname="fanchartdescgramplet.py",
    height=430,
    expand=True,
    gramplet="FanChartDescGramplet",
    detached_height=550,
    detached_width=475,
    gramplet_title=_("Descendant Fan"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="2-Way Fan Chart",
    name=_("2-Way Fan Chart"),
    description=_(
        "Gramplet showing active person's direct ancestors and descendants as a fanchart"
    ),
    status=STABLE,
    fname="fanchart2waygramplet.py",
    height=300,
    expand=True,
    gramplet="FanChart2WayGramplet",
    detached_height=300,
    detached_width=300,
    gramplet_title=_("2-Way Fan"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="FAQ",
    name=_("FAQ"),
    description=_("Gramplet showing frequently asked questions"),
    status=STABLE,
    fname="faqgramplet.py",
    height=300,
    gramplet="FAQGramplet",
    gramplet_title=_("FAQ"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Given Name Cloud",
    name=_("Given Name Cloud"),
    description=_("Gramplet showing all given names as a text cloud"),
    status=STABLE,
    fname="givennamegramplet.py",
    height=300,
    expand=True,
    gramplet="GivenNameCloudGramplet",
    gramplet_title=_("Given Name Cloud"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Pedigree",
    name=_("Pedigree"),
    description=_("Gramplet showing active person's ancestors"),
    status=STABLE,
    fname="pedigreegramplet.py",
    height=300,
    gramplet="PedigreeGramplet",
    gramplet_title=_("Pedigree"),
    expand=True,
    detached_width=600,
    detached_height=400,
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Quick View",
    name=_("Quick View"),
    description=_("Gramplet showing an active item Quick View"),
    status=STABLE,
    fname="quickviewgramplet.py",
    height=300,
    expand=True,
    gramplet="QuickViewGramplet",
    gramplet_title=_("Quick View"),
    detached_width=600,
    detached_height=400,
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Relatives",
    name=_("Relatives"),
    description=_("Gramplet showing active person's relatives"),
    status=STABLE,
    fname="relativegramplet.py",
    height=200,
    gramplet="RelativesGramplet",
    gramplet_title=_("Relatives"),
    detached_width=250,
    detached_height=300,
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Session Log",
    name=_("Session Log"),
    description=_("Gramplet showing all activity for this session"),
    status=STABLE,
    fname="sessionloggramplet.py",
    height=230,
    # data=['no'],
    gramplet="LogGramplet",
    gramplet_title=_("Session Log"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Statistics",
    name=_("Statistics"),
    description=_("Gramplet showing summary data of the Family Tree"),
    status=STABLE,
    fname="statsgramplet.py",
    height=230,
    expand=True,
    gramplet="StatsGramplet",
    gramplet_title=_("Statistics"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Surname Cloud",
    name=_("Surname Cloud"),
    description=_("Gramplet showing all surnames as a text cloud"),
    status=STABLE,
    fname="surnamecloudgramplet.py",
    height=300,
    expand=True,
    gramplet="SurnameCloudGramplet",
    gramplet_title=_("Surname Cloud"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="To Do",
    name=_("To Do", "gramplet"),
    description=_("Gramplet for displaying a To Do list"),
    status=STABLE,
    fname="todogramplet.py",
    height=300,
    expand=True,
    gramplet="ToDoGramplet",
    gramplet_title=_("To Do", "gramplet"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    navtypes=["Dashboard"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Top Surnames",
    name=_("Top Surnames"),
    description=_("Gramplet showing most frequent surnames in this tree"),
    status=STABLE,
    fname="topsurnamesgramplet.py",
    height=230,
    gramplet="TopSurnamesGramplet",
    gramplet_title=_("Top Surnames"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Welcome",
    name=_("Welcome"),
    description=_("Gramplet showing a welcome message"),
    status=STABLE,
    fname="welcomegramplet.py",
    height=300,
    expand=True,
    gramplet="WelcomeGramplet",
    gramplet_title=_("Welcome to Gramps!"),
    version="1.0.1",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="What's Next",
    name=_("What's Next"),
    description=_("Gramplet suggesting items to research"),
    status=STABLE,
    fname="whatsnext.py",
    height=230,
    expand=True,
    gramplet="WhatNextGramplet",
    gramplet_title=_("What's Next?"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    help_url=GRAMPLET_HELP,
)

# ------------------------------------------------------------------------
# Bottombar
# ------------------------------------------------------------------------
register(
    GRAMPLET,
    id="Person Details",
    name=_("Person Details"),
    description=_("Gramplet showing details of a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="persondetails.py",
    height=200,
    gramplet="PersonDetails",
    gramplet_title=_("Details"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Repository Details",
    name=_("Repository Details"),
    description=_("Gramplet showing details of a repository"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="repositorydetails.py",
    height=200,
    gramplet="RepositoryDetails",
    gramplet_title=_("Details"),
    navtypes=["Repository"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place Details",
    name=_("Place Details"),
    description=_("Gramplet showing details of a place"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="placedetails.py",
    height=200,
    gramplet="PlaceDetails",
    gramplet_title=_("Details"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Media Preview",
    name=_("Media Preview"),
    description=_("Gramplet showing a preview of a media object"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="mediapreview.py",
    height=200,
    gramplet="MediaPreview",
    gramplet_title=_("Media Preview"),
    navtypes=["Media"],
    help_url=GRAMPLET_HELP,
)

try:
    from gi import Repository

    repository = Repository.get_default()
    if repository.enumerate_versions("GExiv2"):
        import gi

        gi.require_version("GExiv2", "0.10")
        from gi.repository import GExiv2

        available = True
    else:
        available = False
except (ImportError, ValueError):
    available = False

if available:
    register(
        GRAMPLET,
        id="Metadata Viewer",
        name=_("Image Metadata"),
        description=_("Gramplet showing metadata for a media object"),
        version="1.0.0",
        gramps_target_version=MODULE_VERSION,
        status=STABLE,
        fname="metadataviewer.py",
        height=200,
        gramplet="MetadataViewer",
        gramplet_title=_("Image Metadata"),
        navtypes=["Media"],
        help_url=GRAMPLET_HELP,
    )
else:
    from gramps.gen.config import config

    if not config.get("interface.ignore-gexiv2"):
        from gramps.gen.constfunc import has_display

        if has_display():
            from gramps.gui.dialog import MessageHideDialog
            from gramps.gen.const import URL_WIKISTRING

            gexiv2_dict = {
                "gramps_wiki_build_gexiv2_url": URL_WIKISTRING
                + "GEPS_029:_GTK3-GObject_introspection"
                "_Conversion#GExiv2_for_Image_metadata"
            }
            title = _("GExiv2 module not loaded.")
            message = _(
                "Image metadata functionality will not be available.\n"
                "To build it for Gramps see "
                "%(gramps_wiki_build_gexiv2_url)s" % gexiv2_dict
            )
            if uistate:
                MessageHideDialog(
                    title, message, "interface.ignore-gexiv2", parent=uistate.window
                )
            else:
                MessageHideDialog(title, message, "interface.ignore-gexiv2")

register(
    GRAMPLET,
    id="Person Residence",
    name=_("Person Residence"),
    description=_("Gramplet showing residence events for a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="personresidence.py",
    height=200,
    gramplet="PersonResidence",
    gramplet_title=_("Residence"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person Events",
    name=_("Person Events"),
    description=_("Gramplet showing the events for a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="events.py",
    height=200,
    gramplet="PersonEvents",
    gramplet_title=_("Events"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family Events",
    name=_("Family Events"),
    description=_("Gramplet showing the events for a family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="events.py",
    height=200,
    gramplet="FamilyEvents",
    gramplet_title=_("Events"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person Gallery",
    name=_("Person Gallery"),
    description=_("Gramplet showing media objects for a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="gallery.py",
    height=200,
    gramplet="PersonGallery",
    gramplet_title=_("Gallery"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family Gallery",
    name=_("Family Gallery"),
    description=_("Gramplet showing media objects for a family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="gallery.py",
    height=200,
    gramplet="FamilyGallery",
    gramplet_title=_("Gallery"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Event Gallery",
    name=_("Event Gallery"),
    description=_("Gramplet showing media objects for an event"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="gallery.py",
    height=200,
    gramplet="EventGallery",
    gramplet_title=_("Gallery"),
    navtypes=["Event"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place Gallery",
    name=_("Place Gallery"),
    description=_("Gramplet showing media objects for a place"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="gallery.py",
    height=200,
    gramplet="PlaceGallery",
    gramplet_title=_("Gallery"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Source Gallery",
    name=_("Source Gallery"),
    description=_("Gramplet showing media objects for a source"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="gallery.py",
    height=200,
    gramplet="SourceGallery",
    gramplet_title=_("Gallery"),
    navtypes=["Source"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Citation Gallery",
    name=_("Citation Gallery"),
    description=_("Gramplet showing media objects for a citation"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="gallery.py",
    height=200,
    gramplet="CitationGallery",
    gramplet_title=_("Gallery"),
    navtypes=["Citation"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person Attributes",
    name=_("Person Attributes"),
    description=_("Gramplet showing the attributes of a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="attributes.py",
    height=200,
    gramplet="PersonAttributes",
    gramplet_title=_("Attributes"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Event Attributes",
    name=_("Event Attributes"),
    description=_("Gramplet showing the attributes of an event"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="attributes.py",
    height=200,
    gramplet="EventAttributes",
    gramplet_title=_("Attributes"),
    navtypes=["Event"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family Attributes",
    name=_("Family Attributes"),
    description=_("Gramplet showing the attributes of a family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="attributes.py",
    height=200,
    gramplet="FamilyAttributes",
    gramplet_title=_("Attributes"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Media Attributes",
    name=_("Media Attributes"),
    description=_("Gramplet showing the attributes of a media object"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="attributes.py",
    height=200,
    gramplet="MediaAttributes",
    gramplet_title=_("Attributes"),
    navtypes=["Media"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Source Attributes",
    name=_("Source Attributes"),
    description=_("Gramplet showing the attributes of a source object"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="attributes.py",
    height=200,
    gramplet="SourceAttributes",
    gramplet_title=_("Attributes"),
    navtypes=["Source"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Citation Attributes",
    name=_("Citation Attributes"),
    description=_("Gramplet showing the attributes of a citation object"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="attributes.py",
    height=200,
    gramplet="CitationAttributes",
    gramplet_title=_("Attributes"),
    navtypes=["Citation"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person Notes",
    name=_("Person Notes"),
    description=_("Gramplet showing the notes for a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="PersonNotes",
    gramplet_title=_("Notes"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Event Notes",
    name=_("Event Notes"),
    description=_("Gramplet showing the notes for an event"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="EventNotes",
    gramplet_title=_("Notes"),
    navtypes=["Event"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family Notes",
    name=_("Family Notes"),
    description=_("Gramplet showing the notes for a family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="FamilyNotes",
    gramplet_title=_("Notes"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place Notes",
    name=_("Place Notes"),
    description=_("Gramplet showing the notes for a place"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="PlaceNotes",
    gramplet_title=_("Notes"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Source Notes",
    name=_("Source Notes"),
    description=_("Gramplet showing the notes for a source"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="SourceNotes",
    gramplet_title=_("Notes"),
    navtypes=["Source"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Citation Notes",
    name=_("Citation Notes"),
    description=_("Gramplet showing the notes for a citation"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="CitationNotes",
    gramplet_title=_("Notes"),
    navtypes=["Citation"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Repository Notes",
    name=_("Repository Notes"),
    description=_("Gramplet showing the notes for a repository"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="RepositoryNotes",
    gramplet_title=_("Notes"),
    navtypes=["Repository"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Media Notes",
    name=_("Media Notes"),
    description=_("Gramplet showing the notes for a media object"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="MediaNotes",
    gramplet_title=_("Notes"),
    navtypes=["Media"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Note Notes",
    name=_("Notes"),
    description=_("Gramplet showing the selected note"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notes.py",
    height=200,
    gramplet="NoteNotes",
    gramplet_title=_("Notes"),
    navtypes=["Note"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person Citations",
    name=_("Person Citations"),
    description=_("Gramplet showing the citations for a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="citations.py",
    height=200,
    gramplet="PersonCitations",
    gramplet_title=_("Citations"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Event Citations",
    name=_("Event Citations"),
    description=_("Gramplet showing the citations for an event"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="citations.py",
    height=200,
    gramplet="EventCitations",
    gramplet_title=_("Citations"),
    navtypes=["Event"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family Citations",
    name=_("Family Citations"),
    description=_("Gramplet showing the citations for a family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="citations.py",
    height=200,
    gramplet="FamilyCitations",
    gramplet_title=_("Citations"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place Citations",
    name=_("Place Citations"),
    description=_("Gramplet showing the citations for a place"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="citations.py",
    height=200,
    gramplet="PlaceCitations",
    gramplet_title=_("Citations"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Media Citations",
    name=_("Media Citations"),
    description=_("Gramplet showing the citations for a media object"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="citations.py",
    height=200,
    gramplet="MediaCitations",
    gramplet_title=_("Citations"),
    navtypes=["Media"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person Children",
    name=_("Person Children"),
    description=_("Gramplet showing the children of a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="children.py",
    height=200,
    gramplet="PersonChildren",
    gramplet_title=_("Children"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family Children",
    name=_("Family Children"),
    description=_("Gramplet showing the children of a family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="children.py",
    height=200,
    gramplet="FamilyChildren",
    gramplet_title=_("Children"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person Backlinks",
    name=_("Person References"),
    description=_("Gramplet showing the backlink references for a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="PersonBacklinks",
    gramplet_title=_("References"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Event Backlinks",
    name=_("Event References"),
    description=_("Gramplet showing the backlink references for an event"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="EventBacklinks",
    gramplet_title=_("References"),
    navtypes=["Event"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family Backlinks",
    name=_("Family References"),
    description=_("Gramplet showing the backlink references for a family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="FamilyBacklinks",
    gramplet_title=_("References"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place Backlinks",
    name=_("Place References"),
    description=_("Gramplet showing the backlink references for a place"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="PlaceBacklinks",
    gramplet_title=_("References"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Source Backlinks",
    name=_("Source References"),
    description=_("Gramplet showing the backlink references for a source"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="SourceBacklinks",
    gramplet_title=_("References"),
    navtypes=["Source"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Citation Backlinks",
    name=_("Citation References"),
    description=_("Gramplet showing the backlink references for a citation"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="CitationBacklinks",
    gramplet_title=_("References"),
    navtypes=["Citation"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Repository Backlinks",
    name=_("Repository References"),
    description=_("Gramplet showing the backlink references for a repository"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="RepositoryBacklinks",
    gramplet_title=_("References"),
    navtypes=["Repository"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Media Backlinks",
    name=_("Media References"),
    description=_("Gramplet showing the backlink references for a media object"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="MediaBacklinks",
    gramplet_title=_("References"),
    navtypes=["Media"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Note Backlinks",
    name=_("Note References"),
    description=_("Gramplet showing the backlink references for a note"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="backlinks.py",
    height=200,
    gramplet="NoteBacklinks",
    gramplet_title=_("References"),
    navtypes=["Note"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person Filter",
    name=_("Person Filter"),
    description=_("Gramplet providing a person filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="PersonFilter",
    gramplet_title=_("Filter"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family Filter",
    name=_("Family Filter"),
    description=_("Gramplet providing a family filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="FamilyFilter",
    gramplet_title=_("Filter"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Event Filter",
    name=_("Event Filter"),
    description=_("Gramplet providing an event filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="EventFilter",
    gramplet_title=_("Filter"),
    navtypes=["Event"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Source Filter",
    name=_("Source Filter"),
    description=_("Gramplet providing a source filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="SourceFilter",
    gramplet_title=_("Filter"),
    navtypes=["Source"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Citation Filter",
    name=_("Citation Filter"),
    description=_("Gramplet providing a citation filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="CitationFilter",
    gramplet_title=_("Filter"),
    navtypes=["Citation"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place Filter",
    name=_("Place Filter"),
    description=_("Gramplet providing a place filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="PlaceFilter",
    gramplet_title=_("Filter"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Media Filter",
    name=_("Media Filter"),
    description=_("Gramplet providing a media filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="MediaFilter",
    gramplet_title=_("Filter"),
    navtypes=["Media"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Repository Filter",
    name=_("Repository Filter"),
    description=_("Gramplet providing a repository filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="RepositoryFilter",
    gramplet_title=_("Filter"),
    navtypes=["Repository"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Note Filter",
    name=_("Note Filter"),
    description=_("Gramplet providing a note filter"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="filter.py",
    height=200,
    gramplet="NoteFilter",
    gramplet_title=_("Filter"),
    navtypes=["Note"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Records Gramplet",
    name=_("Records"),
    description=_("Shows some interesting records about people and families"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="recordsgramplet.py",
    authors=["Reinhard Müller"],
    authors_email=["reinhard.mueller@bytewise.at"],
    gramplet="RecordsGramplet",
    height=230,
    expand=True,
    gramplet_title=_("Records"),
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Person To Do",
    name=_("Person To Do"),
    description=_("Gramplet showing the To Do notes for a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="todo.py",
    height=200,
    gramplet="PersonToDo",
    gramplet_title=_("To Do", "gramplet"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Event To Do",
    name=_("Event To Do"),
    description=_("Gramplet showing the To Do notes for an event"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="todo.py",
    height=200,
    gramplet="EventToDo",
    gramplet_title=_("To Do", "gramplet"),
    navtypes=["Event"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Family To Do",
    name=_("Family To Do"),
    description=_("Gramplet showing the To Do notes for a family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="todo.py",
    height=200,
    gramplet="FamilyToDo",
    gramplet_title=_("To Do", "gramplet"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place To Do",
    name=_("Place To Do"),
    description=_("Gramplet showing the To Do notes for a place"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="todo.py",
    height=200,
    gramplet="PlaceToDo",
    gramplet_title=_("To Do", "gramplet"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Source To Do",
    name=_("Source To Do"),
    description=_("Gramplet showing the To Do notes for a source"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="todo.py",
    height=200,
    gramplet="SourceToDo",
    gramplet_title=_("To Do", "gramplet"),
    navtypes=["Source"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Citation To Do",
    name=_("Citation To Do"),
    description=_("Gramplet showing the To Do notes for a citation"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="todo.py",
    height=200,
    gramplet="CitationToDo",
    gramplet_title=_("To Do", "gramplet"),
    navtypes=["Citation"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Repository To Do",
    name=_("Repository To Do"),
    description=_("Gramplet showing the To Do notes for a repository"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="todo.py",
    height=200,
    gramplet="RepositoryToDo",
    gramplet_title=_("To Do", "gramplet"),
    navtypes=["Repository"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Media To Do",
    name=_("Media To Do"),
    description=_("Gramplet showing the To Do notes for a media object"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="todo.py",
    height=200,
    gramplet="MediaToDo",
    gramplet_title=_("To Do", "gramplet"),
    navtypes=["Media"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Python Evaluation",
    name="Python Evaluation",
    description="Gramplet allowing the evaluation of python code",
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=UNSTABLE,
    fname="eval.py",
    height=200,
    gramplet="PythonEvaluation",
    gramplet_title="Python Evaluation",
    help_url=DEBUG_HELP,
)

register(
    GRAMPLET,
    id="Uncollected Objects",
    name="Uncollected Objects",
    description="Gramplet showing uncollected objects",
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=UNSTABLE,
    fname="leak.py",
    height=400,
    gramplet="Leak",
    gramplet_title="Uncollected Objects",
    help_url=DEBUG_HELP,
)

register(
    GRAMPLET,
    id="SoundEx Generator",
    name=_("SoundEx"),
    description=_("Gramplet to generate SoundEx codes"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="soundgen.py",
    height=80,
    gramplet="SoundGen",
    gramplet_title=_("SoundEx"),
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place Enclosed By",
    name=_("Place Enclosed By"),
    description=_("Gramplet showing the places enclosed by the active place"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="locations.py",
    height=200,
    gramplet="EnclosedBy",
    gramplet_title=_("Enclosed By"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Place Encloses",
    name=_("Place Encloses"),
    description=_("Gramplet showing the places that the active place encloses"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="locations.py",
    height=200,
    gramplet="Encloses",
    gramplet_title=_("Encloses"),
    navtypes=["Place"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Geography coordinates for Person Events",
    name=_("Geography coordinates for Person Events"),
    description=_("Gramplet showing the events for a person"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="coordinates.py",
    height=200,
    gramplet="GeoPersonEvents",
    gramplet_title=_("Events Coordinates"),
    navtypes=["Person"],
    help_url=GRAMPLET_HELP,
)

register(
    GRAMPLET,
    id="Geography coordinates for Family Events",
    name=_("Geography coordinates for Family Events"),
    description=_("Gramplet showing the events for all the family"),
    version="1.0.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="coordinates.py",
    height=200,
    gramplet="GeoFamilyEvents",
    gramplet_title=_("Events Coordinates"),
    navtypes=["Family"],
    help_url=GRAMPLET_HELP,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
# Copyright (C) 2017       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Class handling language-specific selection for date parser and displayer.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".gen.datehandler")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ._dateparser import DateParser
from ._datedisplay import DateDisplay, DateDisplayEn, DateDisplayGB
from ..constfunc import win
from ..const import GRAMPS_LOCALE as glocale
from ..utils.grampslocale import GrampsLocale

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
LANG = glocale.calendar

# If LANG contains ".UTF-8" use only the part to the left of "."
# Otherwise some date handler will not load.
if LANG and ".UTF-8" in LANG.upper():
    LANG = LANG.split(".")[0]

if not LANG:
    if "LANG" in os.environ:
        LANG = os.environ["LANG"]

if LANG:
    LANG_SHORT = LANG.split("_")[0]
else:
    LANG_SHORT = "C"

LANG = str(LANG)
LANG_SHORT = str(LANG_SHORT)

LANG_TO_PARSER = {
    "C": DateParser,
}

LANG_TO_DISPLAY = {
    "C": DateDisplayEn,
    "ko_KR": DateDisplay,
}

main_locale = {}  # this will be augmented by calls to register_datehandler

locale_tformat = {}  # locale "tformat" (date format) strings

for no_handler in (
    ("C", ("%d/%m/%Y",)),
    ("eo_EO", "eo", "Esperanto", ("%d/%m/%Y",)),  # 'eo_EO' is a placeholder
    ("sq_AL", "sq", "Albanian", ("%Y/%b/%d",)),
    ("ta_IN", "ta", "Tamil", ("%A %d %B %Y",)),
    ("tr_TR", "tr", "Turkish", ("%d/%m/%Y",)),
    ("vi_VN", "vi", "Vietnamese", ("%d/%m/%Y",)),
):
    format_string = ""
    for possible_format in no_handler:
        if isinstance(possible_format, tuple):
            format_string = possible_format[0]  # pre-seeded date format string
            # maintain legacy gramps transformations
            format_string = format_string.replace("%y", "%Y").replace("-", "/")
    for lang_str in no_handler:
        if isinstance(lang_str, tuple):
            continue
        main_locale[lang_str] = no_handler[0]
        locale_tformat[lang_str] = format_string  # locale's date format string


def register_datehandler(locales, parse_class, display_class):
    """
    Registers the passed date parser class and date displayer
    classes with the specified language locales.

    Set the parser_class and display_class ._locale attribute
    to the corresponding :class:`.GrampsLocale` object.

    :param locales: tuple of strings containing language codes.
                    The character encoding is not included, so the language
                    should be in the form of fr_FR, not fr_FR.utf8
    :type locales: tuple
    :param parse_class: Class to be associated with parsing
    :type parse_class: :class:`.DateParser`
    :param display_class: Class to be associated with displaying
    :type display_class: :class:`.DateDisplay`
    """
    format_string = ""
    for possible_format in locales:  # allow possibly embedding a date format
        if isinstance(possible_format, tuple):
            format_string = possible_format[0]  # pre-seeded date format string
            # maintain legacy gramps transformations
            format_string = format_string.replace("%y", "%Y").replace("-", "/")
    for lang_str in locales:
        if isinstance(lang_str, tuple):
            continue
        LANG_TO_PARSER[lang_str] = parse_class
        LANG_TO_DISPLAY[lang_str] = display_class
        main_locale[lang_str] = locales[0]
        locale_tformat[lang_str] = format_string  # locale's date format string

    parse_class._locale = display_class._locale = GrampsLocale(lang=locales[0])


register_datehandler(
    ("en_GB", "English_United Kingdom", ("%d/%m/%y",)), DateParser, DateDisplayGB
)

register_datehandler(
    ("en_US", "en", "English_United States", ("%m/%d/%y",)), DateParser, DateDisplayEn
)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2012  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2014       Paul Franklin
# Copyright (C) 2010-2015  Craig J. Anderson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Reports/Graphical Reports/Ancestor Tree"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------

from __future__ import annotations
from typing import Any

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------

from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.errors import ReportError
from gramps.gen.plug.menu import (
    TextOption,
    NumberOption,
    BooleanOption,
    EnumeratedListOption,
    StringOption,
    PersonOption,
)
from gramps.gen.plug.report import Report, MenuReportOptions, stdoptions
from gramps.gen.plug.docgen import (
    FontStyle,
    ParagraphStyle,
    GraphicsStyle,
    FONT_SANS_SERIF,
    PARA_ALIGN_CENTER,
)
from gramps.plugins.lib.libtreebase import *
from gramps.plugins.lib.librecurse import AscendPerson
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.display.name import displayer as _nd

PT2CM = utils.pt2cm
# cm2pt = utils.cm2pt

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------
_BORN = (_("b.", "birth abbreviation"),)
_DIED = (_("d.", "death abbreviation"),)
_MARR = (_("m.", "marriage abbreviation"),)

LVL_GEN, LVL_INDX, LVL_Y = range(3)


# ------------------------------------------------------------------------
#
# Box classes
#
# ------------------------------------------------------------------------
class PersonBox(BoxBase):
    """
    Calculates information about the box that will print on a page
    """

    def __init__(self, level):
        BoxBase.__init__(self)
        self.boxstr = "AC2-box"
        # self.level = (level[0]-1, level[1])
        self.level = level
        self.idx = 0

    def __lt__(self, other):
        return self.level[LVL_Y] < other.level[LVL_Y]

    def display(self):
        BoxBase.display(self)
        if self.idx > 0:
            doc = self.page.canvas.doc
            style_sheet = doc.get_style_sheet()
            yoff = PT2CM(
                style_sheet.get_paragraph_style("AC2-Normal").get_font().get_size()
            )
            xbegin = self.x_cm - self.page.page_x_offset
            ybegin = self.y_cm - self.page.page_y_offset - yoff
            doc.draw_text(self.boxstr, "%d" % self.idx, xbegin, ybegin)


class FamilyBox(BoxBase):
    """
    Calculates information about the box that will print on a page
    """

    def __init__(self, level):
        BoxBase.__init__(self)
        self.boxstr = "AC2-fam-box"
        # self.level = (level[0]-1, level[1])
        self.level = level

    def __lt__(self, other):
        return self.level[LVL_Y] < other.level[LVL_Y]


# ------------------------------------------------------------------------
#
# Titles Class(es)
#
# ------------------------------------------------------------------------
class TitleN(TitleNoDisplay):
    """No Title class for the report"""

    def __init__(self, doc, locale):
        TitleNoDisplay.__init__(self, doc, "AC2-Title-box")
        self._ = locale.translation.sgettext

    def calc_title(self, center):
        """Calculate the title of the report"""
        # we want no text, but need a text for the TOC in a book!
        self.mark_text = self._("Ancestor Graph")
        self.text = ""


class TitleA(TitleBox):
    """Title class for the report"""

    def __init__(self, doc, locale, name_displayer):
        self._nd = name_displayer
        TitleBox.__init__(self, doc, "AC2-Title-box")
        self._ = locale.translation.sgettext

    def calc_title(self, center):
        """Calculate the title of the report"""
        name = ""
        if center is not None:
            name = self._nd.display(center)

        # feature request 2356: avoid genitive form
        self.text = self._("Ancestor Graph for %s") % name
        self.set_box_height_width()


# ------------------------------------------------------------------------
#
# CalcItems           (helper class to calculate text)
# make_ancestor_tree  (main recursive functions)
#
# ------------------------------------------------------------------------
class CalcItems:
    """A helper class to calculate the default box text
    and text for each person / marriage
    """

    def __init__(self, dbase):
        _gui = GUIConnect()
        self._gui = _gui

        # calculate the printed lines for each box
        # str = ""
        # if self.get_val('miss_val'):
        #    str = "_____"
        display_repl = _gui.get_val("replace_list")
        self.__calc_l = CalcLines(dbase, display_repl, _gui.locale, _gui.n_d)

        self.__blank_father = None
        self.__blank_mother = None

        self.__blank_father = self.__calc_l.calc_lines(
            None, None, _gui.get_val("father_disp")
        )
        self.__blank_mother = self.__calc_l.calc_lines(
            None, None, _gui.get_val("mother_disp")
        )

        self.center_use = _gui.get_val("center_uses")
        self.disp_father = _gui.get_val("father_disp")
        self.disp_mother = _gui.get_val("mother_disp")

        self.disp_marr = [_gui.get_val("marr_disp")]
        self.__blank_marriage = self.__calc_l.calc_lines(None, None, self.disp_marr)

    def calc_person(self, index, indi_handle, fams_handle):
        working_lines = ""
        if index[1] % 2 == 0 or (index[1] == 1 and self.center_use == 0):
            if indi_handle == fams_handle is None:
                working_lines = self.__calc_l.calc_lines(
                    None, None, self._gui.get_val("father_disp")
                )
            else:
                working_lines = self.disp_father
        else:
            if indi_handle == fams_handle is None:
                working_lines = self.__calc_l.calc_lines(
                    None, None, self._gui.get_val("mother_disp")
                )
            else:
                working_lines = self.disp_mother

        if indi_handle == fams_handle is None:
            return working_lines
        else:
            return self.__calc_l.calc_lines(indi_handle, fams_handle, working_lines)

    def calc_marriage(self, indi_handle, fams_handle):
        if indi_handle == fams_handle is None:
            return self.__blank_marriage
        else:
            return self.__calc_l.calc_lines(indi_handle, fams_handle, self.disp_marr)


class MakeAncestorTree(AscendPerson):
    """
    The main procedure to use recursion to make the tree based off of a person.
    order of people inserted into Persons is important.
    makes sure that order is done correctly.
    """

    def __init__(self, dbase, canvas):
        _gui = GUIConnect()

        AscendPerson.__init__(self, dbase, _gui.maxgen(), _gui.fill_out())
        self.database = dbase
        self.canvas = canvas
        self.inlc_marr = _gui.inc_marr()
        self.inc_sib = _gui.inc_sib()
        self.compress_tree = _gui.compress_tree()
        self.center_family = None
        self.lines = [None] * (_gui.maxgen() + 1)
        if _gui.get_val("show_idx"):
            self.start_idx = _gui.get_val("start_idx")
        else:
            self.start_idx = -1

        self.max_generation = 0

        self.calc_items = CalcItems(self.database)

    def add_person(self, index, indi_handle, fams_handle):
        """Makes a person box and add that person into the Canvas."""

        # print str(index) + " add_person " + str(indi_handle)
        myself = PersonBox((index[0] - 1,) + index[1:])

        if index[LVL_GEN] == 1:  # Center Person
            self.center_family = fams_handle

        if index[LVL_GEN] > self.max_generation:
            self.max_generation = index[LVL_GEN]

        myself.text = self.calc_items.calc_person(index, indi_handle, fams_handle)
        # myself.text[0] = myself.text[0] + ' ' + repr(index) # for debugging

        if self.start_idx >= 0:
            genIdx = self.start_idx * (2 ** (index[LVL_GEN] - 1))
            genOff = index[LVL_INDX] - (2 ** (index[LVL_GEN] - 1))
            myself.idx = genIdx + genOff
        if indi_handle is not None:  # None is legal for an empty box
            myself.add_mark(
                self.database, self.database.get_person_from_handle(indi_handle)
            )

        self.canvas.add_box(myself)

        # make the lines
        indx = index[LVL_GEN]
        self.lines[indx] = myself
        if indx > 1:
            if self.lines[indx - 1].line_to is None:
                line = LineBase(self.lines[indx - 1])
                self.lines[indx - 1].line_to = line
                self.canvas.add_line(line)
            else:
                line = self.lines[indx - 1].line_to
            line.add_to(myself)

        return myself

    def add_person_again(self, index, indi_handle, fams_handle):
        self.add_person(index, indi_handle, fams_handle)

    def add_marriage(self, index, indi_handle, fams_handle):
        """Makes a marriage box and add that person into the Canvas."""

        if not self.inlc_marr:
            return

        myself = FamilyBox((index[0] - 1,) + index[1:])

        # calculate the text.
        myself.text = self.calc_items.calc_marriage(indi_handle, fams_handle)

        self.canvas.add_box(myself)

    def y_index(self, x_level, index):
        """Calculate the column or generation that this person is in.
        x_level  -> 0 to max_gen-1
        index    -> 1 to (self.max_generation**2)-1
        """
        # Calculate which row in the column of people.
        tmp_y = index - (2**x_level)
        # Calculate which row in the table (yes table) of people.
        delta = (2**self.max_generation) // (2 ** (x_level))
        return int((delta / 2) + (tmp_y * delta)) - 1

    def do_y_indx(self):
        """Make the y_index for all boxes
        first off of a forumula, then remove blank areas around the edges,
        then compress the tree if desired
        """
        min_y = self.y_index(
            self.canvas.boxes[0].level[LVL_GEN], self.canvas.boxes[0].level[LVL_INDX]
        )
        for box in self.canvas.boxes:
            if "fam" in box.boxstr:
                box.level = box.level + (
                    self.y_index(box.level[LVL_GEN] - 1, int(box.level[LVL_INDX] / 2)),
                )
            else:
                box.level = box.level + (
                    self.y_index(box.level[LVL_GEN], box.level[LVL_INDX]),
                )
            min_y = min(min_y, box.level[LVL_Y])
            # print (str(box.level))

        # if a last father (of fathers) does not have a father/parents
        # Then there could be a gap.  Remove this gap
        if min_y > 0:
            for box in self.canvas.boxes:
                box.level = box.level[:LVL_Y] + (box.level[LVL_Y] - min_y,)

        # Now that we have y_index, lets see if we need to squish the tree
        self.canvas.boxes.sort()  # Sort them on the y_index
        if not self.compress_tree:
            return
        # boxes are already in top down [LVL_Y] form so lets
        # set the box in the correct y level depending on compress_tree
        y_level = 0
        current_y = self.canvas.boxes[0].level[LVL_Y]
        for box in self.canvas.boxes:
            y_index = box.level[LVL_Y]
            if y_index > current_y:
                current_y = y_index
                y_level += 1
            box.level = box.level[:LVL_Y] + (y_level,)

    def do_sibs(self):
        if not self.inc_sib or self.center_family is None:
            return

        family = self.database.get_family_from_handle(self.center_family)
        mykids = [kid.ref for kid in family.get_child_ref_list()]

        if len(mykids) == 1:  # No other siblings.  Don't do anything.
            return

        # The first person is the center person had he/she has our information
        center = self.canvas.boxes.pop(self.canvas.boxes.index(self.lines[1]))
        line = center.line_to
        level = center.level[LVL_Y]

        move = level - (len(mykids) // 2) + ((len(mykids) + 1) % 2)

        if move < 0:
            # more kids than parents.  ran off the page.  Move them all down
            for box in self.canvas.boxes:
                box.level = (box.level[0], box.level[1], box.level[2] - move)
            move = 0

        line.start = []
        rrr = -1  # if len(mykids)%2 == 1 else 0
        for kid in mykids:
            rrr += 1
            mee = self.add_person((1, 1, move + rrr), kid, self.center_family)
            line.add_from(mee)
            # mee.level = (0, 1, level - (len(mykids)//2)+rrr)
        mee.line_to = line

    def start(self, person_id):
        """go ahead and make it happen"""
        center = self.database.get_person_from_gramps_id(person_id)
        if center is None:
            raise ReportError(_("Person %s is not in the Database") % person_id)
        center_h = center.get_handle()

        # Step 1.  Get the people
        self.recurse(center_h)

        # Step 2.  Calculate the y_index for everyone
        self.do_y_indx()

        # Step 3.  Siblings of the center person
        self.do_sibs()


# ------------------------------------------------------------------------
#
# Transform Classes
#
# ------------------------------------------------------------------------
# ------------------------------------------------------------------------
# Class lr_Transform
# ------------------------------------------------------------------------
class LRTransform:
    """
    setup all of the boxes on the canvas in for a left/right report
    """

    def __init__(self, canvas, max_generations):
        self.canvas = canvas
        self.rept_opts = canvas.report_opts
        self.y_offset = self.rept_opts.littleoffset * 2 + self.canvas.title.height

    def _place(self, box):
        """put the box in it's correct spot"""
        # 1. cm_x
        box.x_cm = self.rept_opts.littleoffset
        box.x_cm += box.level[LVL_GEN] * (
            self.rept_opts.col_width + self.rept_opts.max_box_width
        )
        # 2. cm_y
        box.y_cm = self.rept_opts.max_box_height + self.rept_opts.box_pgap
        box.y_cm *= box.level[LVL_Y]
        box.y_cm += self.y_offset
        # if box.height < self.rept_opts.max_box_height:
        #    box.y_cm += ((self.rept_opts.max_box_height - box.height) /2)

    def place(self):
        """Step through boxes so they can be put in the right spot"""
        # prime the pump
        self.__last_y_level = self.canvas.boxes[0].level[LVL_Y]
        # go
        for box in self.canvas.boxes:
            self._place(box)


# ------------------------------------------------------------------------
#
# class make_report
#
# ------------------------------------------------------------------------
class MakeReport:
    def __init__(self, dbase, doc, canvas, font_normal):
        self.database = dbase
        self.doc = doc
        self.canvas = canvas
        self.font_normal = font_normal

        _gui = GUIConnect()
        self.inlc_marr = _gui.inc_marr()
        self.compress_tree = _gui.compress_tree()

        self.mother_ht = self.father_ht = 0

        self.max_generations = 0

    def get_height_width(self, box):
        """
        obtain width information for each level (x)
        obtain height information for each item
        """

        self.canvas.set_box_height_width(box)

        if box.width > self.canvas.report_opts.max_box_width:
            self.canvas.report_opts.max_box_width = box.width  # + box.shadow

        if box.level[LVL_Y] > 0:
            if box.level[LVL_INDX] % 2 == 0 and box.height > self.father_ht:
                self.father_ht = box.height
            elif box.level[LVL_INDX] % 2 == 1 and box.height > self.mother_ht:
                self.mother_ht = box.height

        if box.level[LVL_GEN] > self.max_generations:
            self.max_generations = box.level[LVL_GEN]

    def get_generations(self):
        return self.max_generations

    def start(self):
        # __gui = GUIConnect()
        # 1.
        # set the sizes for each box and get the max_generations.
        self.father_ht = 0.0
        self.mother_ht = 0.0
        for box in self.canvas.boxes:
            self.get_height_width(box)

        if self.compress_tree and not self.inlc_marr:
            self.canvas.report_opts.max_box_height = min(self.father_ht, self.mother_ht)
        else:
            self.canvas.report_opts.max_box_height = max(self.father_ht, self.mother_ht)

        # At this point we know everything we need to make the report.
        # Size of each column of people - self.rept_opt.box_width
        # size of each column (or row) of lines - self.rept_opt.col_width
        # size of each row - self.rept_opt.box_height
        # go ahead and set it now.
        for box in self.canvas.boxes:
            box.width = self.canvas.report_opts.max_box_width

        # 2.
        # setup the transform class to move around the boxes on the canvas
        transform = LRTransform(self.canvas, self.max_generations)
        transform.place()


class GUIConnect:
    """This is a BORG object.  There is ONLY one.
    This give some common routines that EVERYONE can use like
      get the value from a GUI variable
    """

    __shared_state: dict[str, Any] = {}

    def __init__(self):  # We are BORG!
        self.__dict__ = self.__shared_state

    def set__opts(self, options, locale, name_displayer):
        """Set only once as we are BORG."""
        self.__opts = options
        self.locale = locale
        self.n_d = name_displayer

    def get_val(self, val):
        """Get a GUI value."""
        value = self.__opts.get_option_by_name(val)
        if value:
            return value.get_value()
        else:
            False

    def title_class(self, doc):
        """Return a class that holds the proper title based off of the
        GUI options"""
        title_type = self.get_val("report_title")
        if title_type:
            return TitleA(doc, self.locale, self.n_d)
        else:
            return TitleN(doc, self.locale)

    def inc_marr(self):
        return self.get_val("inc_marr")

    def inc_sib(self):
        return self.get_val("inc_siblings")

    def maxgen(self):
        return self.get_val("maxgen")

    def fill_out(self):
        return self.get_val("fill_out")

    def compress_tree(self):
        return self.get_val("compress_tree")


# ------------------------------------------------------------------------
#
# AncestorTree
#
# ------------------------------------------------------------------------
class AncestorTree(Report):
    """AncestorTree Report"""

    def __init__(self, database, options, user):
        """
        Create AncestorTree object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance
        """
        Report.__init__(self, database, options, user)

        self.options = options
        self._user = user

        self.set_locale(options.menu.get_option_by_name("trans").get_value())
        stdoptions.run_date_format_option(self, options.menu)
        stdoptions.run_private_data_option(self, options.menu)
        stdoptions.run_living_people_option(self, options.menu, self._locale)
        self.database = CacheProxyDb(self.database)
        stdoptions.run_name_format_option(self, options.menu)
        self._nd = self._name_display

    def begin_report(self):
        """
        This report needs the following parameters (class variables)
        that come in the options class.

        max_generations - Maximum number of generations to include.
        pagebbg      - Whether to include page breaks between generations.
        dispf        - Display format for the output box.
        scale_report - Whether to scale the report to fit the width or all.
        indblank     - Whether to include blank pages.
        compress     - Whether to compress chart.
        incl_private - Whether to include private data
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death

        We will
        1. a canvas in its full one-page size
        2. a page that we wish to print on
        scale up/down either or both of the above as needed/desired.
        almost all of this should be moved into Canvas!
        """

        database = self.database

        self.connect = GUIConnect()
        self.connect.set__opts(self.options.menu, self._locale, self._nd)

        # Set up the canvas that we will print on.
        style_sheet = self.doc.get_style_sheet()
        font_normal = style_sheet.get_paragraph_style("AC2-Normal").get_font()

        # The canvas that we will put our report on and print off of
        self.canvas = Canvas(self.doc, ReportOptions(self.doc, font_normal, "AC2-line"))

        self.canvas.report_opts.box_shadow *= self.connect.get_val("shadowscale")
        self.canvas.report_opts.box_pgap *= self.connect.get_val("box_Yscale")
        self.canvas.report_opts.box_mgap *= self.connect.get_val("box_Yscale")

        with self._user.progress(
            _("Ancestor Tree"), _("Making the Tree..."), 4
        ) as step:
            # make the tree onto the canvas
            # inlc_marr = self.connect.get_val("inc_marr")
            self.max_generations = self.connect.get_val("maxgen")
            tree = MakeAncestorTree(database, self.canvas)
            tree.start(self.connect.get_val("pid"))
            tree = None

            step()

            # Title
            title = self.connect.title_class(self.doc)
            center = self.database.get_person_from_gramps_id(
                self.connect.get_val("pid")
            )
            title.calc_title(center)
            self.canvas.add_title(title)

            # make the report as big as it wants to be.
            report = MakeReport(database, self.doc, self.canvas, font_normal)
            report.start()
            self.max_generations = report.get_generations()  # already know
            report = None

            step()

            # Note?
            if self.connect.get_val("inc_note"):
                note_box = NoteBox(
                    self.doc, "AC2-note-box", self.connect.get_val("note_place")
                )
                subst = SubstKeywords(self.database, self._locale, self._nd, None, None)
                note_box.text = subst.replace_and_clean(
                    self.connect.get_val("note_disp")
                )
                self.canvas.add_note(note_box)

            # Now we have the report in its full size.
            # Do we want to scale the report?
            one_page = self.connect.get_val("resize_page")
            scale_report = self.connect.get_val("scale_tree")

            scale = self.canvas.scale_report(
                one_page, scale_report != 0, scale_report == 2
            )

            step()

            if scale != 1 or self.connect.get_val("shadowscale") != 1.0:
                self.scale_styles(scale)

    def write_report(self):
        one_page = self.connect.get_val("resize_page")
        # scale_report = self.connect.get_val("scale_tree")

        # inlc_marr = self.connect.get_val("inc_marr")
        inc_border = self.connect.get_val("inc_border")
        incblank = self.connect.get_val("inc_blank")
        prnnum = self.connect.get_val("inc_pagenum")

        #####################
        # Setup page information

        colsperpage = self.doc.get_usable_width()
        colsperpage += self.canvas.report_opts.col_width
        colsperpage = int(
            colsperpage
            / (
                self.canvas.report_opts.max_box_width
                + self.canvas.report_opts.col_width
            )
        )
        colsperpage = colsperpage or 1

        #####################
        # Vars
        if prnnum:
            page_num_box = PageNumberBox(self.doc, "AC2-box", self._locale)

        # TODO - Here

        #####################
        # ok, everyone is now ready to print on the canvas.  Paginate?
        self.canvas.paginate(colsperpage, one_page)

        #####################
        # Yeah!!!
        # lets finally make some pages!!!
        #####################
        pages = self.canvas.page_count(incblank)
        with self._user.progress(
            _("Ancestor Tree"), _("Printing the Tree..."), pages
        ) as step:
            for page in self.canvas.page_iter_gen(incblank):
                self.doc.start_page()

                # do we need to print a border?
                if inc_border:
                    page.draw_border("AC2-line")

                # Do we need to print the page number?
                if prnnum:
                    page_num_box.display(page)

                # Print the individual people and lines
                page.display()

                step()
                self.doc.end_page()

    def scale_styles(self, scale):
        """
        Scale the styles for this report.
        """
        style_sheet = self.doc.get_style_sheet()

        graph_style = style_sheet.get_draw_style("AC2-box")
        graph_style.set_shadow(
            graph_style.get_shadow(), self.canvas.report_opts.box_shadow * scale
        )
        graph_style.set_line_width(graph_style.get_line_width() * scale)
        style_sheet.add_draw_style("AC2-box", graph_style)

        graph_style = style_sheet.get_draw_style("AC2-fam-box")
        graph_style.set_shadow(
            graph_style.get_shadow(), self.canvas.report_opts.box_shadow * scale
        )
        graph_style.set_line_width(graph_style.get_line_width() * scale)
        style_sheet.add_draw_style("AC2-fam-box", graph_style)

        graph_style = style_sheet.get_draw_style("AC2-note-box")
        # graph_style.set_shadow(graph_style.get_shadow(),
        #                       self.canvas.report_opts.box_shadow * scale)
        graph_style.set_line_width(graph_style.get_line_width() * scale)
        style_sheet.add_draw_style("AC2-note-box", graph_style)

        para_style = style_sheet.get_paragraph_style("AC2-Normal")
        font = para_style.get_font()
        font.set_size(font.get_size() * scale)
        para_style.set_font(font)
        style_sheet.add_paragraph_style("AC2-Normal", para_style)

        para_style = style_sheet.get_paragraph_style("AC2-Note")
        font = para_style.get_font()
        font.set_size(font.get_size() * scale)
        para_style.set_font(font)
        style_sheet.add_paragraph_style("AC2-Note", para_style)

        para_style = style_sheet.get_paragraph_style("AC2-Title")
        font = para_style.get_font()
        font.set_size(font.get_size() * scale)
        para_style.set_font(font)
        style_sheet.add_paragraph_style("AC2-Title", para_style)

        graph_style = GraphicsStyle()
        width = graph_style.get_line_width()
        width = width * scale
        graph_style.set_line_width(width)
        style_sheet.add_draw_style("AC2-line", graph_style)

        self.doc.set_style_sheet(style_sheet)


# ------------------------------------------------------------------------
#
# AncestorTreeOptions
#
# ------------------------------------------------------------------------
class AncestorTreeOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        self.box_Y_sf = None
        self.box_shadow_sf = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        return _nd.display(person)

    def add_menu_options(self, menu):
        ##################
        category_name = _("Tree Options")

        self.__pid = PersonOption(_("Center Person"))
        self.__pid.set_help(_("The center person for the tree"))
        menu.add_option(category_name, "pid", self.__pid)

        siblings = BooleanOption(_("Include siblings of the center person"), False)
        siblings.set_help(
            _(
                "Whether to only display the center person or all "
                "of his/her siblings too"
            )
        )
        menu.add_option(category_name, "inc_siblings", siblings)

        self.max_gen = NumberOption(_("Generations"), 10, 1, 50)
        self.max_gen.set_help(_("The number of generations to include " "in the tree"))
        menu.add_option(category_name, "maxgen", self.max_gen)

        self.fillout = EnumeratedListOption(_("Display unknown\ngenerations"), 0)
        self.fillout.set_help(
            _("The number of generations of empty " "boxes that will be displayed")
        )
        menu.add_option(category_name, "fill_out", self.fillout)

        self.max_gen.connect("value-changed", self.__fillout_vals)
        self.__fillout_vals()

        compress = BooleanOption(_("Compress tree"), True)
        compress.set_help(
            _(
                "Whether to remove any extra blank spaces set "
                "aside for people that are unknown"
            )
        )
        menu.add_option(category_name, "compress_tree", compress)

        self.show_idx = BooleanOption(_("Show Index"), False)
        self.show_idx.set_help(_("Display index of each person"))
        menu.add_option(category_name, "show_idx", self.show_idx)
        self.show_idx.connect("value-changed", self._showidx_changed)

        self.start_idx = NumberOption(_("Start Index"), 1, 1, 50)
        self.start_idx.set_help(_("The start index"))
        menu.add_option(category_name, "start_idx", self.start_idx)

        # better to 'Show siblings of\nthe center person
        # Spouse_disp = EnumeratedListOption(_("Show spouses of\nthe center "
        #                                     "person"), 0)
        # Spouse_disp.add_item(0, _("No.  Do not show Spouses"))
        # Spouse_disp.add_item(1, _("Yes, and use the Main Display Format"))
        # Spouse_disp.add_item(2, _("Yes, and use the Secondary "
        #                           "Display Format"))
        # Spouse_disp.set_help(_("Show spouses of the center person?"))
        # menu.add_option(category_name, "Spouse_disp", Spouse_disp)

        ##################
        category_name = _("Report Options")

        self.title = EnumeratedListOption(_("Report Title"), 0)
        self.title.add_item(0, _("Do not include a title"))
        self.title.add_item(1, _("Include Report Title"))
        self.title.set_help(_("Choose a title for the report"))
        menu.add_option(category_name, "report_title", self.title)

        border = BooleanOption(_("Include a border"), False)
        border.set_help(_("Whether to make a border around the report."))
        menu.add_option(category_name, "inc_border", border)

        prnnum = BooleanOption(_("Include Page Numbers"), False)
        prnnum.set_help(_("Whether to print page numbers on each page."))
        menu.add_option(category_name, "inc_pagenum", prnnum)

        self.scale = EnumeratedListOption(_("Scale tree to fit"), 0)
        self.scale.add_item(0, _("Do not scale tree"))
        self.scale.add_item(1, _("Scale tree to fit page width only"))
        self.scale.add_item(2, _("Scale tree to fit the size of the page"))
        self.scale.set_help(_("Whether to scale the tree to fit a specific paper size"))
        menu.add_option(category_name, "scale_tree", self.scale)
        self.scale.connect("value-changed", self.__check_blank)

        if "BKI" not in self.name.split(","):
            self.__onepage = BooleanOption(
                _(
                    "Resize Page to Fit Tree size\n"
                    "\n"
                    "Note: Overrides options in the 'Paper Option' tab"
                ),
                False,
            )
            self.__onepage.set_help(
                _(
                    "Whether to resize the page to fit the size \n"
                    "of the tree.  Note:  the page will have a \n"
                    "non standard size.\n"
                    "\n"
                    "With this option selected, the following will happen:\n"
                    "\n"
                    "With the 'Do not scale tree' option the page\n"
                    "  is resized to the height/width of the tree\n"
                    "\n"
                    "With 'Scale tree to fit page width only' the height of\n"
                    "  the page is resized to the height of the tree\n"
                    "\n"
                    "With 'Scale tree to fit the size of the page' the page\n"
                    "  is resized to remove any gap in either height or width"
                )
            )
            menu.add_option(category_name, "resize_page", self.__onepage)
            self.__onepage.connect("value-changed", self.__check_blank)
        else:
            self.__onepage = None

        self.__blank = BooleanOption(_("Include Blank Pages"), True)
        self.__blank.set_help(_("Whether to include pages that are blank."))
        menu.add_option(category_name, "inc_blank", self.__blank)

        self.__check_blank()

        ##################
        category_name = _("Report Options (2)")

        stdoptions.add_name_format_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

        ##################
        category_name = _("Display")

        disp = TextOption(
            _("Father\nDisplay Format"), ["$n", "%s $b" % _BORN, "-{%s $d}" % _DIED]
        )
        disp.set_help(_("Display format for the fathers box."))
        menu.add_option(category_name, "father_disp", disp)

        # Will add when libsubstkeyword supports it.
        # missing = EnumeratedListOption(_("Replace missing\nplaces\\dates \
        #                                 with"), 0)
        # missing.add_item(0, _("Does not display anything"))
        # missing.add_item(1, _("Displays '_____'"))
        # missing.set_help(_("What will print when information is not known"))
        # menu.add_option(category_name, "miss_val", missing)

        disp_mom = TextOption(
            _("Mother\nDisplay Format"),
            ["$n", "%s $b" % _BORN, "%s $m" % _MARR, "-{%s $d}" % _DIED],
        )
        disp_mom.set_help(_("Display format for the mothers box."))
        menu.add_option(category_name, "mother_disp", disp_mom)

        center_disp = EnumeratedListOption(_("Center person uses\n" "which format"), 0)
        center_disp.add_item(0, _("Use Fathers Display format"))
        center_disp.add_item(1, _("Use Mothers display format"))
        center_disp.set_help(_("The display format for the center person"))
        menu.add_option(category_name, "center_uses", center_disp)

        self.incmarr = BooleanOption(_("Include Marriage box"), False)
        self.incmarr.set_help(
            _("Whether to include a separate marital box in the report")
        )
        menu.add_option(category_name, "inc_marr", self.incmarr)
        self.incmarr.connect("value-changed", self._incmarr_changed)

        self.marrdisp = StringOption(_("Marriage\nDisplay Format"), "%s $m" % _MARR)
        self.marrdisp.set_help(_("Display format for the marital box."))
        menu.add_option(category_name, "marr_disp", self.marrdisp)
        self._incmarr_changed()

        ##################
        category_name = _("Advanced")

        repldisp = TextOption(
            _("Replace Display Format:\n'Replace this'/' with this'"), []
        )
        repldisp.set_help(_("i.e.\nUnited States of America/U.S.A."))
        menu.add_option(category_name, "replace_list", repldisp)

        # TODO this code is never used and so I conclude it is for future use
        # self.__include_images = BooleanOption(
        #                       _('Include thumbnail images of people'), False)
        # self.__include_images.set_help(
        #                       _("Whether to include thumbnails of people."))
        # menu.add_option(category_name, "includeImages",
        #                 self.__include_images)

        self.usenote = BooleanOption(_("Include a note"), False)
        self.usenote.set_help(_("Whether to include a note on the report."))
        menu.add_option(category_name, "inc_note", self.usenote)
        self.usenote.connect("value-changed", self._usenote_changed)

        self.notedisp = TextOption(_("Note"), [])
        self.notedisp.set_help(_("Add a note\n\n" "$T inserts today's date"))
        menu.add_option(category_name, "note_disp", self.notedisp)

        locales = NoteType(0, 1)
        self.notelocal = EnumeratedListOption(_("Note Location"), 0)
        for num, text in locales.note_locals():
            self.notelocal.add_item(num, text)
        self.notelocal.set_help(_("Where to place the note."))
        menu.add_option(category_name, "note_place", self.notelocal)
        self._usenote_changed()

        self.box_Y_sf = NumberOption(
            _("inter-box scale factor"), 1.00, 0.10, 2.00, 0.01
        )
        self.box_Y_sf.set_help(_("Make the inter-box spacing bigger or smaller"))
        menu.add_option(category_name, "box_Yscale", self.box_Y_sf)

        self.box_shadow_sf = NumberOption(
            _("box shadow scale factor"), 1.00, 0.00, 2.00, 0.01
        )  # down to 0
        self.box_shadow_sf.set_help(_("Make the box shadow bigger or smaller"))
        menu.add_option(category_name, "shadowscale", self.box_shadow_sf)

    def _incmarr_changed(self):
        """
        If Marriage box is not enabled, disable Marriage Display Format box
        """
        value = self.incmarr.get_value()
        self.marrdisp.set_available(value)

    def _usenote_changed(self):
        """
        If Note box is not enabled, disable Note Location box
        """
        value = self.usenote.get_value()
        self.notelocal.set_available(value)

    def _showidx_changed(self):
        """
        If Show Index box is enabled, enable Start Index field
        """
        value = self.show_idx.get_value()
        self.start_idx.set_available(value)

    def __check_blank(self):
        if self.__onepage:
            value = not self.__onepage.get_value()
        else:
            value = True
        off = value and (self.scale.get_value() != 2)
        self.__blank.set_available(off)

    def __fillout_vals(self):
        max_gen = self.max_gen.get_value()
        old_val = self.fillout.get_value()
        item_list = []
        item_list.append(
            [0, _("No generations of empty boxes " "for unknown ancestors")]
        )
        if max_gen > 1:
            item_list.append(
                [1, _("One Generation of empty boxes " "for unknown ancestors")]
            )

        item_list.extend(
            [itr, str(itr) + _(" Generations of empty boxes for unknown ancestors")]
            for itr in range(2, max_gen)
        )

        self.fillout.set_items(item_list)
        if old_val + 2 > len(item_list):
            self.fillout.set_value(len(item_list) - 2)

    def make_default_style(self, default_style):
        """Make the default output style for the Ancestor Tree."""

        # Paragraph Styles:
        font = FontStyle()
        font.set_size(9)
        font.set_type_face(FONT_SANS_SERIF)
        para_style = ParagraphStyle()
        para_style.set_font(font)
        para_style.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("AC2-Normal", para_style)
        box_shadow = PT2CM(font.get_size()) * 0.6

        font = FontStyle()
        font.set_size(9)
        font.set_type_face(FONT_SANS_SERIF)
        para_style = ParagraphStyle()
        para_style.set_font(font)
        para_style.set_description(_("The basic style used for the note display."))
        default_style.add_paragraph_style("AC2-Note", para_style)

        font = FontStyle()
        font.set_size(16)
        font.set_type_face(FONT_SANS_SERIF)
        para_style = ParagraphStyle()
        para_style.set_font(font)
        para_style.set_alignment(PARA_ALIGN_CENTER)
        para_style.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("AC2-Title", para_style)

        # Draw styles
        graph_style = GraphicsStyle()
        graph_style.set_paragraph_style("AC2-Normal")
        graph_style.set_shadow(1, box_shadow)  # shadow set by text size
        graph_style.set_fill_color((255, 255, 255))
        default_style.add_draw_style("AC2-box", graph_style)

        graph_style = GraphicsStyle()
        graph_style.set_paragraph_style("AC2-Normal")
        # graph_style.set_shadow(0, PT2CM(9))  # shadow set by text size
        graph_style.set_fill_color((255, 255, 255))
        default_style.add_draw_style("AC2-fam-box", graph_style)

        graph_style = GraphicsStyle()
        graph_style.set_paragraph_style("AC2-Note")
        graph_style.set_fill_color((255, 255, 255))
        default_style.add_draw_style("AC2-note-box", graph_style)

        # TODO this seems meaningless, as only the text is displayed
        graph_style = GraphicsStyle()
        graph_style.set_paragraph_style("AC2-Title")
        graph_style.set_color((0, 0, 0))
        graph_style.set_fill_color((255, 255, 255))
        graph_style.set_line_width(0)
        graph_style.set_description(_("Cannot edit this reference"))
        default_style.add_draw_style("AC2-Title-box", graph_style)

        graph_style = GraphicsStyle()
        default_style.add_draw_style("AC2-line", graph_style)


# =====================================
# But even if you should suffer for what is right, you are blessed.
# "Do not fear what they fear ; do not be frightened."
# Take Courage
# 1 Peter 3:14



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# HasFamilyAttribute
#
# -------------------------------------------------------------------------
class HasFamilyAttribute(Rule):
    """Rule that checks for a person with a particular family attribute"""

    labels = [_("Family attribute:"), _("Value:")]
    name = _("People with the family <attribute>")
    description = _("Matches people with the family attribute " "of a particular value")
    category = _("General filters")
    allow_regex = True

    def apply_to_one(self, db: Database, person: Person) -> bool:
        if not self.list[0]:
            return False
        for f_id in person.family_list:
            f = db.get_family_from_handle(f_id)
            if not f:
                continue
            for attr in f.attribute_list:
                if attr:
                    name_match = self.list[0] == attr.type
                    value_match = self.match_substring(1, attr.value)
                    if name_match and value_match:
                        return True
        return False
